using System.Text.Json;
using System.Text.Json.Serialization;

namespace View_by_Distance.Shared.Models;

public class PersonContainer : Properties.IPersonContainer
{

    public int? ApproximateYears { init; get; }
    public Person? Person { init; get; }
    public PersonBirthday[]? Birthdays { init; get; }
    public string[] DisplayDirectoryAllFiles { init; get; }
    public string DisplayDirectoryName { init; get; }
    public long? Key { init; get; }
    public string? PartialKeyFormatted { init; get; }

    [JsonConstructor]
    public PersonContainer(int? approximateYears, Person? person, PersonBirthday[]? birthdays, string[] displayDirectoryAllFiles, string displayDirectoryName, long? key, string? partialKeyFormatted)
    {
        ApproximateYears = approximateYears;
        Person = person;
        Birthdays = birthdays;
        DisplayDirectoryAllFiles = displayDirectoryAllFiles;
        DisplayDirectoryName = displayDirectoryName;
        Key = key;
        PartialKeyFormatted = partialKeyFormatted;
    }

    public PersonContainer(int? approximateYears, string[] displayDirectoryAllFiles, string displayDirectoryName) :
        this(approximateYears, null, null, displayDirectoryAllFiles, displayDirectoryName, null, null)
    { }

    public PersonContainer(int? approximateYears, PersonBirthday birthdays, string displayDirectoryName, long key) :
        this(approximateYears, null, new PersonBirthday[] { birthdays }, Array.Empty<string>(), displayDirectoryName, key, null)
    { }

    public PersonContainer(PersonBirthday birthday, string displayDirectoryName) :
        this(Stateless.Methods.IAge.GetApproximateYears(displayDirectoryName), Stateless.Methods.IPerson.GetPerson(displayDirectoryName, birthday.Value.Ticks, birthday), new PersonBirthday[] { birthday }, Array.Empty<string>(), displayDirectoryName, birthday.Value.Ticks, null)
    { }

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

}